/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.actuators;

import java.io.IOException;
import java.util.LinkedList;
import java.util.Vector;

import de.unibi.airobots.bonsai.core.Actuator;
import de.unibi.airobots.bonsai.util.arm.ArmStatus;
import de.unibi.airobots.btl.data.KatanaGripperData;
import de.unibi.airobots.btl.data.KatanaPoseData;
import de.unibi.airobots.btl.data.PoseData;

public interface PoseActuatorTobi extends Actuator {

    /**
     * Move a Joint of the arm.
     * 
     * @param joint
     * @param value
     * @throws IOException
     */
    void moveJoint(int joint, double value) throws IOException;
    
    /**
     * Executes a grasp movement.
     */
    public void executeGraspMovement () throws IOException;
    
    /**
     * List all angles.
     * 
     * @return joints in a LinkList
     * @throws IOException
     */
    LinkedList<Double> listJoints() throws IOException;

    /**
     * Goto a single Position (fully described by P,N,O,A).
     * 
     * @param px
     * @param py
     * @param pz
     * @param nx
     * @param ny
     * @param nz
     * @param ax
     * @param ay
     * @param az
     * @param ox
     * @param oy
     * @param oz
     * @return <code>true</code> if the position is reachable,
     *         <code>false</code> if not
     * @throws IOException
     */
    boolean goTo(double px, double py, double pz, double nx, double ny,
            double nz, double ax, double ay, double az, double ox, double oy,
            double oz);

    /**
     * List all available Movements.
     * 
     * @return Movements in a LinkedList
     * @throws IOException
     */
    Vector<String> listMovements() throws IOException;

    /**
     * Set a special Movement.
     * 
     * @param name
     * @throws IOException
     */
    void setMovements(String name) throws IOException;
    
    /**
     * Set a special Carry Movement.
     * 
     * @param name
     * @throws IOException
     */
    void setCarryMovements(String name) throws IOException;

    /**
     * Open the Gripper of the arm.
     * 
     * @throws IOException
     */
    void openGripper() throws IOException;

    /**
     * Close the Gripper of the arm.
     */
    void closeGripper() throws IOException;

    /**
     * Open gripper when touching
     * 
     * @throws IOException
     */
    void openGripperWhenTouching(int waitSeconds) throws IOException;

    /**
     * Get the Gripper sensor data
     * 
     * @return sensor data of the gripper
     * @throws IOException
     */
    KatanaGripperData getGipperSensorData() throws IOException;

    /**
     * go to a position
     * 
     * @param position
     *            go to this position
     * @param linear
     *            linear movement or non-linear movement
     */
    boolean goTo(KatanaPoseData position, boolean linear) throws IOException;

    /**
     * execute grasping Strategy
     * 
     * @param targetPose
     *            position of the object
     * @throws IOException
     */
    ArmStatus executeStrategy(PoseData targetPose) throws IOException;

    /**
     * Closing Gripper until a define force is achieved
     * 
     * @throws IOException
     */
    void closeGripperByForce() throws IOException;

    /**
     * Get the current position of the Gripper.
     * @return
     */
    KatanaPoseData getPosition() throws IOException;

}
