/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.actuators;

import java.io.IOException;

import de.unibi.airobots.bonsai.core.Actuator;

/**
 * Interface for actuators controlling a speech synthesizer.
 *
 * @author jwienke
 * @author arothert
 */
public interface SpeechActuator extends Actuator {

    /**
     * Say method with the ability to chose between synchronous and asynchronous
     * speech synthesis.
     *
     * @param text
     *            text to say
     * @param async
     *            if <code>true</code>, the method call will return as fast as
     *            possible, blocks otherwise
     * @throws IOException
     *             communication error
     */
    void say(String text, boolean async) throws IOException;

    /**
     * Synchronous say method. Returns after the complete text was said.
     *
     * @param text
     *            text to say
     * @throws IOException
     *             communication error
     */
    void say(String text) throws IOException;

    /**
     * For self-written prosody for the whole text.
     * @param accented_text
     * @throws IOException
     */
    void sayAccentuated(String accented_text) throws IOException;

    /**
     * For normal text, maybe accentuations within the text.
     * Additionally a configuration file path to set the overall prosody of a text must be given.
     * @param accented_text
     * @param prosodyConfig
     * @throws IOException
     */
    void sayAccentuated(String accented_text, String prosodyConfig) throws IOException;

    /**
     * For self-written prosody for the whole text.
     * @param accented_text
     *          text to say, might invole specific accentuations of parts
     *          of the text (MARYXML - prosody - style).
     * @param async
     *          if <code>true</code>, wait until the complete text was said,
     *          else return as fast as possible from this method call
     * @throws IOException
     */
    void sayAccentuated(String accented_text, boolean async) throws IOException;

    /**
     * For normal text, maybe accentuations within the text.
     * Additionally a configuration file path to set the overall prosody of a text must be given.
     *
     * @param accented_text
     *          text to say, might invole specific accentuations of parts
     *          of the text (MARYXML - prosody - style).
     * @param async
     *          if <code>true</code>, wait until the complete text was said,
     *          else return as fast as possible from this method call
     * @param prosodyConfig
     *          path and file for overall prosody setting of the accented_text
     * @throws IOException
     */
    void sayAccentuated(String accented_text, boolean async, String prosodyConfig) throws IOException;
}
