/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.core.configuration;

import java.io.IOException;
import java.net.URI;
import java.util.Set;

import de.unibi.airobots.bonsai.core.Actuator;
import de.unibi.airobots.bonsai.core.CoreObjectFactory;
import de.unibi.airobots.bonsai.core.Sensor;

/**
 * Parser interface for Bonsai configuration files. The basic procedure is to
 * create an instance of a concrete parser, call the {@link #parseFile(URI)}
 * method on it and collect the results with the various getter methods. If the
 * parsing resulted in an error, this class is in an undefined state and results
 * cannot be used.
 * 
 * A parser instance must provide at least one
 * {@link ConfiguredCoreObjectFactory} describing a {@link CoreObjectFactory}
 * configuration. The described factory is used to create {@link Sensor} and
 * {@link Actuator} instances. Moreover two sets of {@link ConfiguredSensor}s
 * and {@link ConfiguredActuator}s need to be returned. All described
 * {@link Sensor}s and {@link Actuator}s in these sets must be constructible by
 * one of the factories described earlier.
 * 
 * @author jwienke
 */
public interface ConfigurationParser {

    /**
     * Exception thrown for parsing errors.
     * 
     * @author jwienke
     */
    public class ParseException extends RuntimeException {

        /**
         * Serial version UID.
         */
        private static final long serialVersionUID = 2656964892576461162L;

        /**
         * Constructor.
         * 
         * @param message
         *            exception description
         */
        public ParseException(String message) {
            super(message);
        }

        /**
         * Constructor.
         * 
         * @param message
         *            exception description
         * @param cause
         *            exception cause
         */
        public ParseException(String message, Throwable cause) {
            super(message, cause);
        }

        /**
         * Constructor.
         * 
         * @param cause
         *            exception cause
         */
        public ParseException(Throwable cause) {
            super(cause);
        }

    }

    /**
     * After parsing this method returns the list of factories and their options
     * parsed from the configuration file.
     * 
     * @return list of configured factories
     * @throws IllegalStateException
     *             thrown if parsing didn't succeed before this call
     */
    Set<ConfiguredCoreObjectFactory> getConfiguredFactories()
            throws IllegalStateException;

    /**
     * After parsing this method returns the list of sensors and their options
     * parsed from the configuration file.
     * 
     * @return list of configured sensors found in the configuration
     * @throws IllegalStateException
     *             thrown if parsing didn't succeed before this call
     */
    Set<ConfiguredSensor> getConfiguredSensors() throws IllegalStateException;

    /**
     * After parsing this method returns the list of actuators and their options
     * parsed from the configuration file.
     * 
     * @return list of configured actuators found in the configuration
     * @throws IllegalStateException
     *             thrown if parsing didn't succeed before this call
     */
    Set<ConfiguredActuator> getConfiguredActuators()
            throws IllegalStateException;

    /**
     * Parses the file under the given url as configurations. If an exception is
     * thrown, no configuration was parsed.
     * 
     * @param configurationUrl
     *            url of the configuration
     * @throws IOException
     *             error opening the file for parsing
     * @throws ParseException
     *             error parsing the file
     * @throws IllegalStateException
     *             if parsing was already invoked
     */
    void parseFile(URI configurationUrl) throws IOException, ParseException,
            IllegalStateException;

}
