package de.unibi.airobots.bonsai.core.configuration;

import java.io.IOException;
import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.ParsingException;
import nu.xom.ValidityException;

import org.apache.log4j.Logger;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * Parser for XML configuration files. A schema is installed with BonSAI
 * specifying the expected format.
 * 
 * <p>
 * The expected format defines, that all sensors must be defined before any
 * actuator is specified.
 * </p>
 * 
 * @author jwienke
 */
public class XmlConfigurationParser implements ConfigurationParser {

    private Logger logger = Logger.getLogger(getClass());

    private static final String ROOT_ELEMENT_NAME = "BonsaiConfiguration";
    private static final String FACTORY_OPTIONS_ELEMENT_NAME = "FactoryOptions";
    private static final String OPTION_ELEMENT_NAME = "Option";
    private static final String SENSOR_ELEMENT_NAME = "Sensor";
    private static final String OPTIONS_ELEMENT_NAME = "Options";
    private static final String FACTORY_CLASS_ARG = "factoryClass";
    private static final String OPTION_KEY_ARG = "key";
    private static final String SENSOR_KEY_ARG = "key";
    private static final String DATA_TYPE_CLASS_ARG = "dataTypeClass";
    private static final String SENSOR_CLASS_ARG = "sensorClass";
    private static final String ACTUATOR_ELEMENT_NAME = "Actuator";
    private static final String ACTUATOR_KEY_ARG = "key";
    private static final String ACTUATOR_CLASS_ARG = "actuatorClass";
    private static final String ACTUATOR_INTERFACE_ARG = "actuatorInterface";

    private Map<String, ConfiguredCoreObjectFactory> factoriesByClass =
            new HashMap<String, ConfiguredCoreObjectFactory>();
    private Map<String, ConfiguredSensor> sensorsByKey =
            new HashMap<String, ConfiguredSensor>();
    private Map<String, ConfiguredActuator> actuatorsByKey =
            new HashMap<String, ConfiguredActuator>();

    private boolean parsingCompleted = false;
    private boolean parsingError = false;

    /**
     * Constructor.
     */
    public XmlConfigurationParser() {
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Set<ConfiguredCoreObjectFactory> getConfiguredFactories()
            throws IllegalStateException {
        if (!parsingCompleted || parsingError) {
            throw new IllegalStateException(
                    "Parsing was not invoked or failed. "
                            + "No results can be fetched.");
        }
        return new HashSet<ConfiguredCoreObjectFactory>(factoriesByClass
                .values());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Set<ConfiguredSensor> getConfiguredSensors()
            throws IllegalStateException {
        if (!parsingCompleted || parsingError) {
            throw new IllegalStateException(
                    "Parsing was not invoked or failed. "
                            + "No results can be fetched.");
        }
        return new HashSet<ConfiguredSensor>(sensorsByKey.values());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Set<ConfiguredActuator> getConfiguredActuators()
            throws IllegalStateException {
        if (!parsingCompleted || parsingError) {
            throw new IllegalStateException(
                    "Parsing was not invoked or failed. "
                            + "No results can be fetched.");
        }
        return new HashSet<ConfiguredActuator>(actuatorsByKey.values());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void parseFile(URI configurationUrl) throws IOException,
            ParseException, IllegalStateException {

        if (parsingCompleted || parsingError) {
            throw new IllegalStateException("Parser already parsed a file");
        }

        try {
            XMLReader xerces =
                    XMLReaderFactory
                            .createXMLReader("org.apache.xerces.parsers.SAXParser");
            xerces.setFeature(
                    "http://apache.org/xml/features/validation/schema", true);

            Builder parser = new Builder(xerces, true);

            Document doc = parser.build(configurationUrl.toString());

            parseFactorySettings(doc);
            parseSensors(doc);
            parseActuators(doc);

            parsingCompleted = true;

        } catch (ValidityException e) {
            throw new ParseException(e);
        } catch (ParsingException e) {
            throw new ParseException(e);
        } catch (SAXException e) {
            throw new ParseException("Maybe xerces SAXParser missing?", e);
        } finally {
            if (!parsingCompleted) {
                parsingError = true;
            }
        }

    }

    private void parseFactorySettings(Document doc) throws ParseException {

        logger.debug("Parsing factory settings");

        Nodes factoryOptions =
                doc.query("/" + ROOT_ELEMENT_NAME + "/"
                        + FACTORY_OPTIONS_ELEMENT_NAME);
        for (int i = 0; i < factoryOptions.size(); i++) {

            Node factoryOptionNode = factoryOptions.get(i);
            logger.debug("Possible " + FACTORY_OPTIONS_ELEMENT_NAME + ": "
                    + factoryOptionNode.toXML());
            if (!(factoryOptionNode instanceof Element)) {
                logger.warn("Skipping because no "
                        + Element.class.getSimpleName());
                continue;
            }
            Element factoryOption = (Element) factoryOptionNode;

            ConfiguredCoreObjectFactory factory =
                    new ConfiguredCoreObjectFactory();

            // class name
            String factoryClassName =
                    factoryOption.getAttributeValue(FACTORY_CLASS_ARG);
            factory.setFactoryClassName(getTrimmedValue(factoryClassName));

            // options
            Map<String, String> options = parseOptions(factoryOption);
            factory.setFactroyOptions(options);

            // check that a configuration for this element doesn't already exist
            if (factoriesByClass.containsKey(factory.getFactoryClassName())) {
                throw new ParseException(
                        "There is already a configuration for factory class "
                                + factory.getFactoryClassName());
            }
            factoriesByClass.put(factory.getFactoryClassName(), factory);

        }

    }

    private void parseSensors(Document doc) throws ParseException {

        logger.debug("Parsing sensors");

        Nodes sensorNodes =
                doc.query("/" + ROOT_ELEMENT_NAME + "/" + SENSOR_ELEMENT_NAME);
        for (int i = 0; i < sensorNodes.size(); i++) {

            Node sensorNode = sensorNodes.get(i);
            logger.debug("Possible " + SENSOR_ELEMENT_NAME + ": "
                    + sensorNode.toXML());
            if (!(sensorNode instanceof Element)) {
                logger.warn("Skipping because no "
                        + Element.class.getSimpleName());
                continue;
            }
            Element sensor = (Element) sensorNode;

            ConfiguredSensor configuredSensor = new ConfiguredSensor();

            // key
            String key = sensor.getAttributeValue(SENSOR_KEY_ARG);
            configuredSensor.setKey(getTrimmedValue(key));

            // data type class
            String dataTypeClass =
                    sensor.getAttributeValue(DATA_TYPE_CLASS_ARG);
            configuredSensor.setDataType(getTrimmedValue(dataTypeClass));

            // factory class
            String factoryClass = sensor.getAttributeValue(FACTORY_CLASS_ARG);
            configuredSensor.setFactoryClassName(getTrimmedValue(factoryClass));

            // sensor class
            String sensorClass = sensor.getAttributeValue(SENSOR_CLASS_ARG);
            configuredSensor.setSensorClassName(getTrimmedValue(sensorClass));

            // options
            Element optionElement =
                    sensor.getFirstChildElement(OPTIONS_ELEMENT_NAME);
            if (optionElement == null) {
                throw new ParseException(SENSOR_ELEMENT_NAME
                        + " does not contain an " + OPTIONS_ELEMENT_NAME + ": "
                        + sensor.toXML());
            }
            configuredSensor.setOptions(parseOptions(optionElement));

            // check that a sensor with this key doesn't already exist
            if (sensorsByKey.containsKey(configuredSensor.getKey())) {
                throw new ParseException("There is already a sensor for key "
                        + configuredSensor.getKey());
            }
            sensorsByKey.put(configuredSensor.getKey(), configuredSensor);

        }

    }

    private void parseActuators(Document doc) throws ParseException {

        logger.debug("Parsing actuators");

        Nodes actuatorNodes =
                doc
                        .query("/" + ROOT_ELEMENT_NAME + "/"
                                + ACTUATOR_ELEMENT_NAME);
        for (int i = 0; i < actuatorNodes.size(); i++) {

            Node actuatorNode = actuatorNodes.get(i);
            logger.debug("Possible " + ACTUATOR_ELEMENT_NAME + ": "
                    + actuatorNode.toXML());
            if (!(actuatorNode instanceof Element)) {
                logger.warn("Skipping because no "
                        + Element.class.getSimpleName());
                continue;
            }
            Element sensor = (Element) actuatorNode;

            ConfiguredActuator configuredActuator = new ConfiguredActuator();

            // key
            String key = sensor.getAttributeValue(ACTUATOR_KEY_ARG);
            configuredActuator.setKey(getTrimmedValue(key));

            // actuator class
            String actuatorClass = sensor.getAttributeValue(ACTUATOR_CLASS_ARG);
            configuredActuator
                    .setActuatorClassName(getTrimmedValue(actuatorClass));

            // factory class
            String factoryClass = sensor.getAttributeValue(FACTORY_CLASS_ARG);
            configuredActuator
                    .setFactoryClassName(getTrimmedValue(factoryClass));

            // interface
            String interfaceClass =
                    sensor.getAttributeValue(ACTUATOR_INTERFACE_ARG);
            configuredActuator
                    .setInterfaceClassName(getTrimmedValue(interfaceClass));

            // options
            Element optionElement =
                    sensor.getFirstChildElement(OPTIONS_ELEMENT_NAME);
            if (optionElement == null) {
                throw new ParseException(SENSOR_ELEMENT_NAME
                        + " does not contain an " + OPTIONS_ELEMENT_NAME + ": "
                        + sensor.toXML());
            }
            configuredActuator.setOptions(parseOptions(optionElement));

            // check that a sensor with this key doesn't already exist
            if (actuatorsByKey.containsKey(configuredActuator.getKey())) {
                throw new ParseException("There is already a actuator for key "
                        + configuredActuator.getKey());
            }
            actuatorsByKey.put(configuredActuator.getKey(), configuredActuator);

        }

    }

    private Map<String, String> parseOptions(Element optionRoot)
            throws ParseException {

        logger.debug("Parsing options in root: " + optionRoot.toXML());

        Map<String, String> options = new HashMap<String, String>();

        Nodes optionNodes = optionRoot.query(OPTION_ELEMENT_NAME);
        for (int i = 0; i < optionNodes.size(); i++) {

            Node optionNode = optionNodes.get(i);
            logger.debug("Possible " + OPTION_ELEMENT_NAME + ": "
                    + optionNode.toXML());
            if (!(optionNode instanceof Element)) {
                logger.warn("Skipping because no "
                        + Element.class.getSimpleName());
                continue;
            }
            Element option = (Element) optionNode;

            String key =
                    getTrimmedValue(option.getAttributeValue(OPTION_KEY_ARG));
            String value = getTrimmedValue(option.getValue());

            options.put(key, value);

        }

        return options;

    }

    /**
     * Returns a trimmed value of the given string and throws
     * {@link ParseException} for every error that can occur. This may be a
     * <code>null</code> value or an empty string.
     * 
     * @param rawValue
     *            value to trim, may be <code>null</code>
     * @return trimmed string value, not <code>null</code>
     * @throws ParseException
     *             empty string or <code>null</code> value
     */
    private String getTrimmedValue(String rawValue) throws ParseException {

        if (rawValue == null) {
            throw new ParseException("Null value");
        }
        String finalValue = rawValue.trim();
        if (finalValue.isEmpty()) {
            throw new ParseException("Empty value");
        }

        return finalValue;

    }

}
