package de.unibi.airobots.bonsai.skills;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.scxml.Context;
import org.apache.commons.scxml.Evaluator;
import org.apache.commons.scxml.SCXMLExecutor;
import org.apache.commons.scxml.SCXMLListener;
import org.apache.commons.scxml.TriggerEvent;
import org.apache.commons.scxml.env.SimpleContext;
import org.apache.commons.scxml.env.SimpleDispatcher;
import org.apache.commons.scxml.env.SimpleErrorReporter;
import org.apache.commons.scxml.env.jexl.JexlContext;
import org.apache.commons.scxml.env.jexl.JexlEvaluator;
import org.apache.commons.scxml.model.ModelException;
import org.apache.commons.scxml.model.SCXML;
import org.apache.commons.scxml.model.State;
import org.apache.commons.scxml.model.Transition;
import org.apache.commons.scxml.model.TransitionTarget;
import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.core.Actuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.skills.SkillStateMachine.EntryListener;
import de.unibi.airobots.bonsai.skills.state.AbstractState;

/**
 * <p>
 * This class demonstrates one approach for providing the base functionality
 * needed by classes representing stateful entities, whose behaviors are defined
 * via SCXML documents.
 * </p>
 * 
 * <p>
 * SCXML documents (more generically, UML state chart diagrams) can be used to
 * define stateful behavior of objects, and Commons SCXML enables developers to
 * use this model directly into the corresponding code artifacts. The resulting
 * artifacts tend to be much simpler, embody a useful separation of concerns and
 * are easier to understand and maintain. As the size of the modeled entity
 * grows, these benefits become more apparent.
 * </p>
 * 
 * <p>
 * This approach functions by registering an SCXMLListener that gets notified
 * onentry, and calls the namesake method for each state that has been entered.
 * </p>
 * 
 * <p>
 * This class swallows all exceptions only to log them. Developers of subclasses
 * should think of themselves as &quot;component developers&quot; catering to
 * other end users, and therefore ensure that the subclasses are free of
 * <code>ModelException</code>s and the like. Most methods are
 * <code>protected</code> for ease of subclassing.
 * </p>
 * 
 */
public class SkillStateMachine {

    /**
     * The state machine that will drive the instances of this class.
     */
    private SCXML stateMachine;

    /**
     * The instance specific SCXML engine.
     */
    private SCXMLExecutor engine;

    /**
     * 
     */
    private SCXMLSensorActuatorManager saManager;

    /**
     * The log.
     */
    private static final Logger LOG = Logger.getLogger(SkillStateMachine.class);

    /**
     * The method signature for the activities corresponding to each state in
     * the SCXML document.
     */
    private static final Class[] SIGNATURE = new Class[0];

    /**
     * The method parameters for the activities corresponding to each state in
     * the SCXML document.
     */
    private static final Object[] PARAMETERS = new Object[0];

    /**
     * Convenience constructor.
     * 
     * @param stateMachine
     *            The parsed SCXML instance that describes the
     *            &quot;lifecycle&quot; of the instances of this class.
     * 
     * @since 0.7
     */
    public SkillStateMachine(final SCXML stateMachine,
            SCXMLSensorActuatorManager saManager) {

        this(stateMachine, saManager, new JexlContext(), new JexlEvaluator());
    }

    /**
     * Convenience constructor.
     * 
     * @param aXml
     * @param saManager
     */
    public SkillStateMachine(String aXml, SCXMLSensorActuatorManager saManager) {
        SCXMLDecoder scxmlDekoder = new SCXMLDecoder();
        SCXML stateMachine =
                scxmlDekoder.parseAndValidateSCXML2(SkillStateMachine.class
                        .getResource(aXml).toString());

        initialize(stateMachine, new SimpleContext(), new JexlEvaluator());
    }

    /**
     * Primary constructor.
     * 
     * @param stateMachine
     *            The parsed SCXML instance that describes the
     *            &quot;lifecycle&quot; of the instances of this class.
     * @param rootCtx
     *            The root context for this instance.
     * @param evaluator
     *            The expression evaluator for this instance.
     * @param saManager
     * 
     * @see Context
     * @see Evaluator
     */
    public SkillStateMachine(final SCXML stateMachine,
            SCXMLSensorActuatorManager saManager, JexlContext rootCtx,
            JexlEvaluator evaluator) {
        this.saManager = saManager;
        initialize(stateMachine, rootCtx, evaluator);
    }

    /**
     * Instantiate and initialize the underlying executor instance.
     * 
     * @param stateMachine
     *            The state machine
     * @param rootCtx
     *            The root context
     * @param evaluator
     *            The expression evaluator
     */
    private void initialize(final SCXML stateMachine, final Context rootCtx,
            final Evaluator evaluator) {
        engine =
                new SCXMLExecutor(evaluator, new SimpleDispatcher(),
                        new SimpleErrorReporter());
        engine.setStateMachine(stateMachine);
        engine.setSuperStep(true);
        engine.setRootContext(rootCtx);
        engine.addListener(stateMachine, new EntryListener());
        // try {
        // engine.go();
        // } catch (ModelException me) {
        // logError(me);
        // }
    }

    /**
     * Fire an event on the SCXML engine.
     * 
     * @param event
     *            The event name.
     * @return Whether the state machine has reached a &quot;final&quot;
     *         configuration.
     */
    public boolean fireEvent(final String event) {
        // System.out.println("\t\ttriggering: " + event + " = " + this);
        TriggerEvent[] evts =
                { new TriggerEvent(event, TriggerEvent.SIGNAL_EVENT, null) };
        try {
            engine.triggerEvents(evts);
        } catch (ModelException me) {
            logError(me);
        }
        return engine.getCurrentStatus().isFinal();
    }

    /**
     * Get the SCXML object representing this state machine.
     * 
     * @return Returns the stateMachine.
     * @deprecated Returns null, use getEngine().getStateMachine() instead
     */
    public static SCXML getStateMachine() {
        return null;
    }

    /**
     * Get the SCXML engine driving the &quot;lifecycle&quot; of the instances
     * of this class.
     * 
     * @return Returns the engine.
     */
    public SCXMLExecutor getEngine() {
        return engine;
    }

    /**
     * Get the log for this class.
     * 
     * @return Returns the log.
     */
    public Logger getLog() {
        System.out.println("get logger called - investigate!");
        return LOG;
    }

    /**
     * Set the log for this class.
     * 
     * @param log
     *            The log to set.
     */
    public void setLog(final Logger log) {
        System.out.println("set logger called - investigate!");
        // this.log = log;
    }

    /**
     * Invoke the no argument method with the following name.
     * 
     * @param className
     *            The method to invoke.
     * @return Whether the invoke was successful.
     */
    public boolean invoke(final String className) {
        LOG.info("Invoking state/class: " + className);
        try {
            Class<?> c =
                    Class.forName("de.unibi.airobots.bonsai.skills.state."
                            + className);

            Constructor<?>[] constrs = c.getConstructors();
            Constructor<?> aConstructor = constrs[0];

            AbstractState aState;
            aState = (AbstractState) aConstructor.newInstance(this);

            aState.start();

            return true;
        } catch (InstantiationException e) {
            LOG.error("Error while starting a State!: " + e.getMessage());
        } catch (ClassNotFoundException e1) {
            e1.printStackTrace();
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        }
        return false;
    }

    /**
     * Starts the state machine.
     * 
     */
    void startMachine() {
        try {
            engine.go();
        } catch (ModelException me) {
            logError(me);
        }
    }

    /**
     * Reset the state machine.
     * 
     * @return Whether the reset was successful.
     */
    public boolean resetMachine() {
        try {
            engine.reset();
        } catch (ModelException me) {
            logError(me);
            return false;
        }
        return true;
    }

    /**
     * Utility method for logging error.
     * 
     * @param exception
     *            The exception leading to this error condition.
     */
    protected void logError(final Exception exception) {
        LOG.error(exception.getMessage(), exception);
    }

    /**
     * A SCXMLListener that is only concerned about &quot;onentry&quot;
     * notifications.
     */
    protected class EntryListener implements SCXMLListener {

        /**
         * {@inheritDoc}
         */
        public void onEntry(final TransitionTarget entered) {
            invoke(entered.getId());
        }

        /**
         * No-op.
         * 
         * @param from
         *            The &quot;source&quot; transition target.
         * @param to
         *            The &quot;destination&quot; transition target.
         * @param transition
         *            The transition being followed.
         */
        public void onTransition(final TransitionTarget from,
                final TransitionTarget to, final Transition transition) {
            // nothing to do
        }

        /**
         * No-op.
         * 
         * @param exited
         *            The transition target being exited.
         */
        public void onExit(final TransitionTarget exited) {
            // nothing to do
        }

    }

    /**
     * Returns a list of currently possible events.
     * 
     * @return list of Strings with names of the possible events
     */
    public LinkedList<String> getEventList() {
        LinkedList<String> list = new LinkedList<String>();
        List<Transition> a =
                ((State) getEngine().getCurrentStatus().getStates().iterator()
                        .next()).getTransitionsList();
        for (Iterator<Transition> iterator = a.iterator(); iterator.hasNext();) {
            String aString = iterator.next().getEvent().toString();
            list.push(aString);
        }
        return list;
    }

    // /////////////////////////////////////////////////////////////////

    /**
     * Returns the instance of a sensor which is managed in the
     * SCXMLSensorActuatorManager.
     * 
     * @param sensorName
     *            Name of the sensor you want
     * @param aDataTypeClass
     *            Class of the data type the sensor delivers
     * @return the requested sensor
     */
    public <T> Sensor<T> getManagedSensor(String sensorName, Class aDataTypeClass) {
        return saManager.getManagedSensor(sensorName, aDataTypeClass);
    }

    /**
     * Returns the instance of an actuator which is managed in the
     * SCXMLSensorActuatorManager.
     * 
     * @param actuatorName
     *            Name of the actuator you want
     * @param aDataTypeClass
     *            Class of the data type the actuator requires
     * @return the requested actuator
     */
    public <T extends Actuator> T getManagedActuator(String actuatorName, Class<T> aDataTypeClass) {
        return saManager.getManagedActuator(actuatorName, aDataTypeClass);
    }
}
