/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 * 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 * 				Marian Pohling, Lukas Twardon
 */
package de.unibi.airobots.bonsai.skills.state;

import java.io.IOException;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.util.CoordinateSystemConverter;
import de.unibi.airobots.bonsai.util.helper.Movements;
import de.unibi.airobots.btl.data.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.SpeedData;
import de.unibi.airobots.btl.data.NavigationGoalData.GoalType;
import de.unibi.airobots.btl.data.PositionData;

/**
 * Implementation of a simple follow person state.
 * 
 * @author nkoester
 */
public class IntroduceRobot extends AbstractState {

    private static final String PRESS_LATER = "You may press them later.";
    private static final String EMERGENCY_BUTTONS =
            "I also have emergency buttons on both sides.";
    private static final String SENSORS_AND_ACTUATORS =
            "I have several sensors and actuators to explore and interact with the environment.";
    private static final String FIND_MY_WAY =
            "I can use prior maps to find my way, but also adapt to new environments easily.";
    private static final String CHALLENGE =
            "The major challenge for me is to learn about people and their homes by interaction.";
    private Logger logger = Logger.getLogger(this.getClass());

    private SpeechActuator speechActuator;
    private Sensor<GlobalPlannerStateChange> gplanChangeSensor;
    private NavigationActuator navActuator;
    private Sensor<PositionData> positionSensor;
    private Sensor<SpeedData> speedSensor;

    public IntroduceRobot(SkillStateMachine aStateMachine) {
        super(aStateMachine);
    }

    @Override
    public void getRequiredSensorsAndActuators() {
        navActuator =
                BonsaiManager.getInstance().createActuator(
                        "NavigationActuator", NavigationActuator.class);
        positionSensor =
                BonsaiManager.getInstance().createSensor("PositionSensor",
                        PositionData.class);
        gplanChangeSensor =
                BonsaiManager.getInstance().createSensor(
                        "GlobalPlannerStateChangeSensor",
                        GlobalPlannerStateChange.class);
        speedSensor =
                BonsaiManager.getInstance().createSensor("SpeedSensor",
                        SpeedData.class);
        this.speechActuator =
                BonsaiManager.getInstance().createActuator("SpeechActuator",
                        SpeechActuator.class);
    }

    @Override
    public boolean init() {
        System.out.println("\t" + this.getClass().getSimpleName().toString()
                + " init");

        try {
            speechActuator.sayAccentuated(CHALLENGE);
            speechActuator.sayAccentuated(FIND_MY_WAY);
            speechActuator.sayAccentuated(SENSORS_AND_ACTUATORS);

        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        try {

            Thread.sleep(1000);

            PositionData positionData;
            try {
                positionData = positionSensor.readLast();

                NavigationGoalData rotateStart =
                        new NavigationGoalData(positionData);
                rotateStart =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                positionData, (Math.PI / 2), 0.0);

                rotateStart.setYawTolerance(Math.PI * 1 / 8);
                rotateStart.setCoordinateTolerance(0.5);
                try {
                    driveToPosition(rotateStart, GoalType.LOCAL);
                    speechActuator.sayAccentuated(EMERGENCY_BUTTONS);
                    speechActuator.sayAccentuated(PRESS_LATER);
                } catch (IOException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }

                try {
                    positionData = positionSensor.readLast();
                } catch (IOException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
                NavigationGoalData rotateEnd;
                rotateEnd =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                positionData, (-Math.PI / 2), 0.0);

                rotateEnd.setYawTolerance(Math.PI / 1 / 8);
                rotateEnd.setCoordinateTolerance(0.5);
                try {
                    driveToPosition(rotateEnd, GoalType.LOCAL);
                } catch (IOException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            } catch (IOException e1) {
                // TODO Auto-generated catch block
                e1.printStackTrace();
            }
        } catch (InterruptedException e) {
            logger.error("Interrupted during IntroduceYourself");
        }

        return true;
    }

    @Override
    public boolean execute() {

        System.out.println("\t" + this.getClass().getSimpleName().toString()
                + " execute");

        return true;
    }

    @Override
    public boolean end() {

        System.out.println("\t" + this.getClass().getSimpleName().toString()
                + " end");

        return true;
    }

    public boolean driveToPosition(NavigationGoalData target,GoalType type)
            throws IOException {
        return Movements.driveToPosition(navActuator, positionSensor,
                speedSensor, gplanChangeSensor, target, type);
    }

}
