/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.util.helper;

import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.FaceIdentificationSensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.util.CoordinateSystemConverter;
import de.unibi.airobots.btl.data.FaceIdentificationList;
import de.unibi.airobots.btl.data.PersonData;
import de.unibi.airobots.btl.data.PositionData;
import de.unibi.airobots.btl.data.FaceIdentificationList.FaceIdentification;

/**
 * A class responsible for the various management activities related to persons
 * found by the robot.
 * 
 * @author jwienke
 */
public class PersonManager {

	private Logger logger = Logger.getLogger(this.getClass());

	private static final long RECOGNITION_TIME = 20000;

	private static final long MAX_PERCEPTION_AGE = 2000;

	/**
	 * The distance used to recognize known persons by their position in the
	 * world.
	 */
	public static final double PERSON_DISTANCE = 0.5;

	/**
	 * Max. distance to accept a person standing in front of the robot.
	 */
	public static double FRONT_MAX_DISTANCE = 1500;
	
	public static double DISTANCE = 1;

	/**
	 * Max. angle from the front orientation of the robot in which a person is
	 * accepted as standing in front of Tobi. this angle is meant as an absolute
	 * value to each direction.
	 */
	public static double FRONT_MAX_ANGLE = Math.PI / 3.0;

	private int numPersonsIntroduced;
	private int numPersonsFind;
	private int numPersonsRecognize;

	private Sensor<PersonList> personSensor;

	/**
	 * A map of all persons tobi has found while exploring the arena. Each
	 * person is associated with its position in world coordinates.
	 */
	private List<PersonDescription> knownPersons = new LinkedList<PersonDescription>();

	private Set<PersonData> recognizedPersons = new HashSet<PersonData>();

	/**
	 * Constructs a new {@link PersonManager}.
	 */
	public PersonManager(int numPersonsIntroduced,
			int numPersonsRecognize, Sensor<PersonList> personSensor) {
		this.numPersonsIntroduced = numPersonsIntroduced;
		this.numPersonsRecognize = numPersonsRecognize;

		this.personSensor = personSensor;
	}

	/**
	 * Tries to tell if the robot knows a person by using the persons position
	 * in the world. This only works if the persons don't move.
	 * 
	 * @param person
	 *            the person to check knowledge of for
	 * @param currentPosition
	 *            current position of the robot
	 * @return <code>true</code> if the robot thinks he knows the person, else
	 *         <code>false</code>
	 */
	public boolean knowsPerson(final PersonData person,
			final PositionData currentPosition) {

		// lziegler: fixed mm -> m
		NavigationGoalData personInWC = CoordinateSystemConverter
				.polar2NavigationGoalData(currentPosition, person.getAngle(),
						person.getDistance() / 1000.0);

		for (PersonDescription knownPerson : knownPersons) {

			PositionData knownPersonData = knownPerson.getLastPosition();

			if (Math.sqrt(Math.pow(knownPersonData.getX() - personInWC.getX(),
					2)
					+ Math.pow(knownPersonData.getY() - personInWC.getY(), 2)) < PERSON_DISTANCE) {
				return true;
			}

		}

		return false;

	}

	/**
	 * Tries to tell if the robot has recognized a person before by using the
	 * persons position in the world. This only works if the persons don't move.
	 * 
	 * @param person
	 *            the person to check knowledge of for
	 * @param currentPosition
	 *            current position of the robot
	 * @return <code>true</code> if the robot thinks he knows the person, else
	 *         <code>false</code>
	 */
	public boolean recognizedPerson(final PersonData person,
			final PositionData currentPosition) {

		// lziegler: fixed mm -> m
		NavigationGoalData personInWC = CoordinateSystemConverter
				.polar2NavigationGoalData(currentPosition, person.getAngle(),
						person.getDistance() / 1000.0);
		// personInWC [m]

		for (PersonData recognizedPerson : recognizedPersons) {

			PositionData recognizedPersonData = recognizedPerson
					.getPositionData();
			// recognizedPersonData [m]

			logger.debug("regperson: " + recognizedPersonData.getX() + " " + recognizedPersonData.getY());		
			
			if (Math.sqrt(Math.pow(recognizedPersonData.getX()
					- personInWC.getX(), 2)
					+ Math.pow(recognizedPersonData.getY() - personInWC.getY(),
							2)) < PERSON_DISTANCE) {
				return true;
			}

		}

		return false;

	}
	
	/**
	 * Stores this person in the list of known persons.
	 * 
	 * @param person
	 *            the person to store
	 */
	public void storeNewPerson(PersonDescription person) {
		knownPersons.add(person);
	}

	/**
	 * Returns the number of persons the robot already knows.
	 * 
	 * @return positive integer
	 */
	public int getNumberOfKnownPersons() {
		return knownPersons.size();
	}

	/**
	 * Tells if the robot has seen enough persons within introduction.
	 * 
	 * @return <code>true</code> if enough persons were introduced, else
	 *         <code>false</code>
	 */
	public boolean hasSeenEnoughPersons() {
		return getNumberOfKnownPersons() >= numPersonsIntroduced;
	}

	/**
	 * Stores this person in the list of already recognized persons.
	 * 
	 * @param person
	 *            the person that was recognized
	 */
	public void personRecognized(PersonData person) {
		recognizedPersons.add(person);
	}

	/**
	 * Returns the number of recognized persons.
	 * 
	 * @return positive integer
	 */
	public int getRecognizedPersons() {
		return recognizedPersons.size();
	}

	/**
	 * Tells if there have been enough recognitions to fulfill the task.
	 * 
	 * @return <code>true</code> if there were enough recognitions, else
	 *         <code>false</code>
	 */
	public boolean hasRecognizedEnoughPersons() {
		return (getRecognizedPersons() >= numPersonsRecognize);
	}

	/**
	 * Tells if the given person is standing in front of the robot so that it
	 * can be recognized using the camera.
	 * 
	 * @param p
	 *            person to check
	 * @return <code>true</code> if person is standing in front of the robot
	 *         facing it within a small distance, else <code>false</code>
	 */
	public boolean isStandingInFront(PersonData p) {

		// TODO use facing information
		
		//boolean isYoung = System.currentTimeMillis()
		//- p.getTimestamp().getUpdated().getTime() < MAX_PERCEPTION_AGE;
		boolean isSmallId = p.getId() >= 0;
		
		//if (!isYoung) {
		//	logger.debug("perception too old");
		//	return false;
		//}
		
		if (!isSmallId) {
			logger.debug("illegal id: " + p.getId());
			return false;
		}

		if (p.getDistance() > FRONT_MAX_DISTANCE) {
			logger.debug("Person too far away");
			return false;
		}

		if (Math.cos(p.getAngle() / 180.0 * Math.PI) < Math.cos(FRONT_MAX_ANGLE)) {
			logger.debug("Person not in right angle");
			return false;
		}

		// TODO: Wieder einkommentieren, wenn der Person Tracker gefixt ist
		/*
		 * if (p.isWalking()) { logger.debug("Person is walking"); return false;
		 * }
		 */
		logger.debug("Person is in front");
		
		return true;

	}

	/**
	 * Returns the unrecognized person who's distance is the smallest to the
	 * robot.
	 * 
	 * @param currentPosition
	 *            current position to search from
	 * @return unknown person to the robot with smallest distance of all unknown
	 *         persons
	 * @throws IllegalStateException
	 *             no person found
	 * @throws InterruptedException
	 */
	public PersonData getNextUnrecognizedPerson(PositionData currentPosition)
			throws IllegalStateException, InterruptedException {

		logger.debug("Searching unrecognized persons");

		try {

			PersonList persons = personSensor.readLast();
			logger.debug("Found " + persons.size() + " unrecognized persons");

			List<PersonData> unrecognizedPersons = new LinkedList<PersonData>();
			for (PersonData p : persons) {
				//boolean isYoung = System.currentTimeMillis()
				//		- p.getTimestamp().getUpdated().getTime() < MAX_PERCEPTION_AGE;
				boolean isSmallId = p.getId() >= 0;
				if (isSmallId
						&& !recognizedPerson(p, currentPosition)) {
					unrecognizedPersons.add(p);
				}

			}

			if (unrecognizedPersons.isEmpty()) {
				throw new IllegalStateException("No unrecognized persons found");
			} else {
				sortPersonsByDistance(unrecognizedPersons);
				logger.debug(unrecognizedPersons.get(0)+" "+unrecognizedPersons.size());
			}

			return unrecognizedPersons.get(0);

		} catch (IOException e) {
			logger.warn("IOException reading persons", e);
			throw new IllegalStateException("No persons can be read", e);
		}

	}

	/**
	 * Returns the nearest person standing in front of the robot that probably
	 * was not recognized before.
	 * 
	 * @return person
	 * @throws IllegalStateException
	 *             no such person found
	 * @throws InterruptedException
	 */
	public PersonData getNextUnrecognizedPersonInFront()
			throws IllegalStateException, InterruptedException {

		// XXX duplicate code

		try {

			PersonList persons = personSensor.readLast();

			if (persons.size() == 0) {
				throw new IllegalStateException("No persons found.");
			}

			List<PersonData> personList = new LinkedList<PersonData>();
			for (PersonData p : persons) {
				//boolean isYoung = System.currentTimeMillis()
				//		- p.getTimestamp().getUpdated().getTime() < MAX_PERCEPTION_AGE;
				boolean isSmallId = p.getId() >= 0;
				if (isSmallId && isStandingInFront(p)) {
					personList.add(p);
				}
			}
			sortPersonsByDistance(personList);

			if (personList.size() == 0) {
				throw new IllegalStateException(
						"No unrecognized person in front of robot found.");
			}

			return personList.get(0);

		} catch (IOException e) {
			logger.warn("IOError", e);
			throw new IllegalStateException("No persons found (IOException)", e);
		}

	}

	/**
	 * Returns the nearest person in front of the robot that is facing the robot
	 * ready to be recognized or anything else.
	 * 
	 * @return person in front of robot
	 * @throws IllegalStateException
	 *             no such person found
	 * @throws InterruptedException
	 */
	public PersonData getNextPersonInFront() throws IllegalStateException,
			InterruptedException {

		try {

			PersonList persons = personSensor.readLast();

			if (persons.size() == 0) {
				throw new IllegalStateException("No persons found.");
			}

			List<PersonData> personList = new LinkedList<PersonData>();
			for (PersonData p : persons) {
				//boolean isYoung = System.currentTimeMillis()
				//		- p.getTimestamp().getUpdated().getTime() < MAX_PERCEPTION_AGE;
				boolean isSmallId = p.getId() >= 0;
				if (isSmallId && isStandingInFront(p)) {
					personList.add(p);
				}
			}
			sortPersonsByDistance(personList);

			if (personList.size() == 0) {
				throw new IllegalArgumentException(
						"No person in front of robot found.");
			}

			return personList.get(0);

		} catch (IOException e) {
			logger.warn("IOError", e);
			throw new IllegalStateException("No persons found (IOException)", e);
		}

	}

	/**
	 * Sorts a list of persons by distance in place, so that the person with the
	 * smallest distance to the robot is the first element of the list.
	 * 
	 * @param persons
	 *            list of persons to sort
	 */
	private void sortPersonsByDistance(List<PersonData> persons) {
		Collections.sort(persons, new Comparator<PersonData>() {

			@Override
			public int compare(PersonData o1, PersonData o2) {
				return new Double(o1.getDistance()).compareTo(o2.getDistance());
			}

		});
	}

	/**
	 * Getter for the {@link PersonDescription} corresponding to the given id.
	 * Attention: Returns <code>null</code>, if no person corresponds to this
	 * id.
	 * 
	 * @param id
	 *            The id of the person.
	 * @return The {@link PersonDescription} corresponding to the given id or
	 *         <code>null</code>, if no person corresponds to this id.
	 */
	private PersonDescription getPersonDescriptionById(int id) {
		for (PersonDescription person : knownPersons) {
			if (person.getId() == id)
				return person;
		}
		return null;
	}

	/**
	 * Tries to recognize a previously stored person und returns the
	 * corresponding {@link PersonDescription}. Attention: Returns
	 * <code>null</code>, if no person was recognized.
	 * 
	 * @return The {@link PersonDescription} corresponding to the recognized
	 *         PersonData or <code>null</code>, if no person was recognized.
	 * @throws IOException
	 */
	public PersonDescription recognize(FaceIdentificationSensor faceSensor)
			throws IOException {

		faceSensor.activate();

		long startTime = System.currentTimeMillis();
		long maxTime = startTime + RECOGNITION_TIME;
		Map<Integer, Integer> recognitionCounts = new HashMap<Integer, Integer>();

		// collect recognition results
		while (System.currentTimeMillis() < maxTime) {
			long maxWaitTime = maxTime - System.currentTimeMillis();
			if (maxWaitTime < 0) {
				break;
			}
			try {
				
				FaceIdentificationList results = faceSensor
						.readLast(maxWaitTime);

				if (results == null) {
					continue;
				}

				// iterate over all faces classified
				for (FaceIdentification ident : results) {
					if (ident.isIdentified()) {

						// count how often each id was founds
						if (!recognitionCounts.containsKey(ident.getClassId())) {
							recognitionCounts.put(ident.getClassId(), 0);
						}
						recognitionCounts.put(ident.getClassId(),
								recognitionCounts.get(ident.getClassId()) + 1);

					}
				}

			} catch (InterruptedException e) {
				logger.warn("Ignoring interruption", e);
				continue;
			}
		}

		faceSensor.deactivate();

		// evaluate results
		if (recognitionCounts.isEmpty()) {
			// no one recognized
			return null;
		}

		int maxCount = -1;
		int maxClassId = -1;
		for (Entry<Integer, Integer> classCount : recognitionCounts.entrySet()) {
			if (classCount.getValue() > maxCount) {
				maxCount = classCount.getValue();
				maxClassId = classCount.getKey();
			}
		}

		assert maxClassId >= 0;
		
		System.out.println("PersonManager.recognize: " + getPersonDescriptionById(maxClassId));
		
		return getPersonDescriptionById(maxClassId);

	}

	public PersonDescription getKnownPerson(int id) {
		return knownPersons.get(id);
	}
}
