/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.util.slam;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.LinkedList;

import javax.swing.AbstractButton;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.border.Border;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.filechooser.FileFilter;
import javax.swing.tree.DefaultMutableTreeNode;

import nu.xom.Element;
import nu.xom.Elements;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.util.Pair;
import de.unibi.airobots.bonsai.util.helper.annotation.MapArea;
import de.unibi.airobots.bonsai.util.helper.annotation.MapPoint;
import de.unibi.airobots.bonsai.util.helper.annotation.MapRegion;
import de.unibi.airobots.bonsai.util.slam.AnnotationDataTreePanel.AnnotationChangeListener;
import de.unibi.airobots.bonsai.util.slam.gui.MapAnnotationPanel;
import de.unibi.airobots.bonsai.util.slam.gui.MapAnnotationPanel.WorldClickListener;
import de.unibi.airobots.bonsai.util.xml.ParameterMissingException;
import de.unibi.airobots.bonsai.util.xml.XMLDecoderUtil;
import de.unibi.airobots.bonsai.util.xml.XMLEncoderUtil;
import de.unibi.airobots.btl.data.BinarySlamMap;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.PositionData;
import de.unibi.airobots.btl.data.NavigationGoalData.GoalType;

public class SlamAnnotationTools {

    private static final String CURRENT_POINT = "current Point";

    private static final String LOAD = "load";

    private static final String ANNOTATION = "annotation";

    private static final String OTHER = "OTHER";

    private static final String SAVE = "save";

    private static final int DEFAULT_UPDATE_TIME = 500;

    /** For logging output. */
    private static final Logger LOGGER =
            Logger.getLogger(SlamAnnotationTools.class);

    private static final String AREA = "AREA";

    private static final String VIEW_POINT = "view_point";

    private static MapRegion mapRegion;

    private static ButtonGroup areaButtons;

    private static JTextField labelText;

    private static HashMap<String, MapArea> areas =
            new HashMap<String, MapArea>();

    private static JTextField arenaText;

    private static JTextField groupText;

    private static ButtonGroup annotationButtons;

    private static AnnotationDataTreePanel annotationTreePanel;

    private static MapAnnotationPanel mapPanel;

    private static JRadioButton annotationRadioButton;

    private static JRadioButton viewPointRadioButton;

    private static boolean edit = false;

    private static Sensor<PositionData> positionSensor;

    private static NavigationActuator navActuator;

    private static JTextField heightText;

    public static void showMapAnnotationWindow(
            Sensor<BinarySlamMap> slamSensor,
            Sensor<PositionData> posSensor,
            NavigationActuator navigationActuator,
            final String annotationXmlPath) {

        positionSensor = posSensor;

        navActuator = navigationActuator;

        // Create the map panel
        mapPanel = new MapAnnotationPanel(slamSensor, DEFAULT_UPDATE_TIME);

        MapAnnotationPanel.WorldClickListener navigator =
                createWorldClickListener(mapPanel);

        mapPanel.addWorldClickListener(navigator);

        ActionListener listener = new ActionListener() {

            @Override
            public void actionPerformed(ActionEvent e) {

                String action = e.getActionCommand();

                File f = new File("/vol/robocup/trunk/etc/SlamAnnotation");
                LOGGER.debug("filechooser path " + f.getAbsolutePath());
                LOGGER.debug("path exists " + f.exists());
                JFileChooser fc;
                if (f.exists()) {
                    fc = new JFileChooser(f);
                } else {
                    fc = new JFileChooser();
                }

                fc.setFileFilter(new FileFilter() {

                    @Override
                    public boolean accept(File f) {
                        return f.getName().toLowerCase().endsWith(".xml");
                    }

                    @Override
                    public String getDescription() {
                        return "xml";
                    }
                });

                if (SAVE.equals(action)) {
                    fc.setDialogType(JFileChooser.SAVE_DIALOG);

                    int state = fc.showOpenDialog(null);
                    if (state == JFileChooser.APPROVE_OPTION) {
                        File file = fc.getSelectedFile();
                        saveXML(file.getAbsolutePath());
                        LOGGER.debug(file.getAbsolutePath());
                    }

                } else if (LOAD.equals(action)) {

                    fc.setDialogType(JFileChooser.OPEN_DIALOG);

                    int state = fc.showOpenDialog(null);
                    if (state == JFileChooser.APPROVE_OPTION) {
                        File file = fc.getSelectedFile();
                        LOGGER.debug(file.getAbsolutePath());

                        try {
                            Element root =
                                    XMLDecoderUtil.loadXML(file
                                            .getAbsolutePath());

                            Elements child = root.getChildElements();

                            for (int i = 0; i < child.size(); i++) {
                                try {
                                    MapArea mapArea =
                                            XMLDecoderUtil.getMapArea(child
                                                    .get(i));
                                    annotationTreePanel.addAnnotation(mapArea);
                                    areas
                                            .put(mapArea.getType().name(),
                                                    mapArea);
                                } catch (NumberFormatException e1) {
                                    LOGGER.warn(e1);
                                } catch (ParameterMissingException e1) {
                                    LOGGER.warn(e1);
                                }
                            }

                            updateGuiComponents();

                        } catch (Exception e1) {
                            LOGGER.warn(e1);
                        }
                    }
                } else if (CURRENT_POINT.equals(action)) {
                    try {
                        PositionData pos = positionSensor.readLast();

                        NavigationGoalData nav = new NavigationGoalData(pos);
                        nav.setType(GoalType.GLOBAL);

                        LinkedList<MapPoint> points =
                                new LinkedList<MapPoint>();

                        MapPoint p1 =
                                new MapPoint(pos.getX() - 1, pos.getY() - 1,
                                        MapPoint.Kind.WORLD, MapPoint.Unit.M,
                                        "");
                        MapPoint p2 =
                                new MapPoint(pos.getX() - 1, pos.getY() + 1,
                                        MapPoint.Kind.WORLD, MapPoint.Unit.M,
                                        "");
                        MapPoint p3 =
                                new MapPoint(pos.getX() + 1, pos.getY() + 1,
                                        MapPoint.Kind.WORLD, MapPoint.Unit.M,
                                        "");
                        MapPoint p4 =
                                new MapPoint(pos.getX() + 1, pos.getY() - 1,
                                        MapPoint.Kind.WORLD, MapPoint.Unit.M,
                                        "");

                        points.add(p1);
                        points.add(p2);
                        points.add(p3);
                        points.add(p4);

                        String areaLabel =
                                areaButtons.getSelection().getActionCommand();

                        String arenaName = arenaText.getText();
                        double height = Double.parseDouble(heightText.getText());

                        String regionGroup = groupText.getText();
                        String label = labelText.getText();

                        MapRegion regio =
                                new MapRegion(MapRegion.Rectangle_TYPE, label,
                                        regionGroup, points, nav, height);

                        if (!areas.containsKey(areaLabel)) {
                            MapArea mapArea =
                                    new MapArea(MapArea.MapAreaType.valueOf(areaLabel));
                            mapArea.addRegion(regio);
                            mapArea.setArenaName(arenaName);
                            areas.put(areaLabel, mapArea);
                        } else {
                            regio.setLabel(label);
                            regio.setGroup(regionGroup);
                            areas.get(areaLabel).addRegion(regio);
                        }

                        updateGuiComponents();
                    } catch (IOException e1) {
                        LOGGER.warn(e1);
                    } catch (InterruptedException e1) {
                        LOGGER.warn(e1);
                    }
                }
            }
        };

        JMenuBar menubar = new JMenuBar();

        JButton saveItem = new JButton(SAVE);
        saveItem.addActionListener(listener);

        JButton loadItem = new JButton(LOAD);
        loadItem.addActionListener(listener);

        JButton currenPointItem = new JButton(CURRENT_POINT);
        currenPointItem.addActionListener(listener);

        menubar.add(saveItem);
        menubar.add(loadItem);
        menubar.add(currenPointItem);

        areaButtons = new ButtonGroup();

        MapArea.MapAreaType[] areaTyps = MapArea.MapAreaType.values();
        int typCount = areaTyps.length;

        JPanel typPanel = new JPanel(new GridLayout(typCount, 1));

        for (int i = 0; i < typCount; i++) {

            JRadioButton button = new JRadioButton(areaTyps[i].name());
            if (i == 0)
                button.setSelected(true);
            button.setActionCommand(areaTyps[i].name());
            areaButtons.add(button);
            typPanel.add(button);

        }

        JPanel leftPanel = new JPanel(new GridLayout(2, 1));
        leftPanel.add(typPanel);

        annotationButtons = new ButtonGroup();

        JPanel annotationPanel = new JPanel(new GridLayout(2, 1));
        annotationRadioButton = new JRadioButton(ANNOTATION);
        annotationRadioButton.setActionCommand(ANNOTATION);
        annotationRadioButton.setSelected(true);
        annotationButtons.add(annotationRadioButton);
        annotationPanel.add(annotationRadioButton);

        viewPointRadioButton = new JRadioButton(VIEW_POINT);
        viewPointRadioButton.setActionCommand(VIEW_POINT);
        annotationButtons.add(viewPointRadioButton);
        annotationPanel.add(viewPointRadioButton);

        leftPanel.add(annotationPanel);

        annotationTreePanel = new AnnotationDataTreePanel();
        AnnotationChangeListener annotationListener =
                createAnnotationTreeListener();
        annotationTreePanel.addListener(annotationListener);

        JPanel panel = new JPanel(new BorderLayout(5, 5));
        panel.add(annotationTreePanel, BorderLayout.EAST);
        panel.add(mapPanel, BorderLayout.CENTER);
        panel.add(leftPanel, BorderLayout.WEST);

        TitledBorder title;
        Border loweredetched =
                BorderFactory.createEtchedBorder(EtchedBorder.LOWERED);

        JPanel labelPanel = new JPanel(new GridLayout(2, 2));

        title = BorderFactory.createTitledBorder(loweredetched, "arena");

        JPanel borderPanel = new JPanel(new GridLayout(1, 1));

        arenaText = new JTextField();
        arenaText.setText("arena");

        borderPanel.add(arenaText);
        borderPanel.setBorder(title);

        labelPanel.add(borderPanel);

        title = BorderFactory.createTitledBorder(loweredetched, "group");

        borderPanel = new JPanel(new GridLayout(1, 1));

        groupText = new JTextField();
        groupText.setText("group");

        borderPanel.add(groupText);
        borderPanel.setBorder(title);

        labelPanel.add(borderPanel);

        title = BorderFactory.createTitledBorder(loweredetched, "label");

        borderPanel = new JPanel(new GridLayout(1, 1));

        labelText = new JTextField();
        labelText.setText(OTHER);

        borderPanel.add(labelText);
        borderPanel.setBorder(title);

        labelPanel.add(borderPanel);
        
        title = BorderFactory.createTitledBorder(loweredetched, "height");

        borderPanel = new JPanel(new GridLayout(1, 1));

        heightText = new JTextField();
        heightText.setText("0");

        borderPanel.add(heightText);
        borderPanel.setBorder(title);

        labelPanel.add(borderPanel);

        panel.add(labelPanel, BorderLayout.SOUTH);

        // Create and show a window
        JFrame testFrame = new JFrame("SLAM MapAnnotationPanel");
        testFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        testFrame.setJMenuBar(menubar);
        testFrame.add(panel);

        testFrame.pack();

        testFrame.setVisible(true);

    }

    private static AnnotationChangeListener createAnnotationTreeListener() {
        AnnotationChangeListener listener = new AnnotationChangeListener() {

            @Override
            public void annotationChange(String event,
                    DefaultMutableTreeNode node) {

                if (event.equals(AnnotationDataTreePanel.REMOVE2)) {

                    if (!node.isRoot()) {

                        DefaultMutableTreeNode parent =
                                (DefaultMutableTreeNode) node.getParent();

                        if (parent.isRoot()) {
                            // area
                            String areaType = node.getUserObject().toString();
                            areas.remove(areaType);
                        } else if (!parent.isRoot() && !node.isLeaf()) {

                            // region
                            String areaType = node.getParent().toString();
                            String regionLabel =
                                    node.getUserObject().toString();

                            MapArea area = areas.get(areaType);
                            area.removeRegion(regionLabel);
                        } else if (node.isLeaf()) {
                            // point
                            String areaType =
                                    node.getParent().getParent().toString();
                            String regionLabel = node.getParent().toString();

                            MapArea area = areas.get(areaType);
                            MapRegion region =
                                    area.getRegionByLabel(regionLabel);

                            String[] p =
                                    node.getUserObject().toString().split(" ");

                            if (p[0]
                                    .equals(XMLEncoderUtil.VIEW_POINT_ELEMENTNAME)) {
                                region.setViewPoint(null);
                            } else if (p[0]
                                    .equals(XMLEncoderUtil.POINT_ATTRNAME)) {
                                double x = Double.parseDouble(p[1]);
                                double y = Double.parseDouble(p[2]);

                                region.removePoint(x, y);
                            }
                        }
                    }

                } else if (event.equals(AnnotationDataTreePanel.EDIT2)) {
                    if (!node.isRoot()) {

                        edit = true;
                        DefaultMutableTreeNode parent =
                                (DefaultMutableTreeNode) node.getParent();

                        if (!parent.isRoot() && !node.isLeaf()) {
                            // region
                            String areaType = node.getParent().toString();
                            String regionLabel =
                                    node.getUserObject().toString();

                            MapArea area = areas.get(areaType);
                            mapRegion = area.getRegionByLabel(regionLabel);
                            annotationButtons.setSelected(annotationRadioButton
                                    .getModel(), true);
                            for (Enumeration<AbstractButton> it =
                                    areaButtons.getElements(); it
                                    .hasMoreElements();) {
                                AbstractButton b = it.nextElement();
                                String action = b.getActionCommand();

                                if (action.equals(areaType)) {
                                    areaButtons.setSelected(b.getModel(), true);
                                    break;
                                }

                            }
                            arenaText.setText(area.getArenaName());
                            groupText.setText(mapRegion.getGroup());
                            labelText.setText(mapRegion.getLabel());
                            heightText.setText(String.valueOf(mapRegion.getHeight()));

                        } else if (node.isLeaf()) {
                            // point
                            String areaType =
                                    node.getParent().getParent().toString();
                            String regionLabel = node.getParent().toString();

                            MapArea area = areas.get(areaType);
                            mapRegion = area.getRegionByLabel(regionLabel);

                            String[] p =
                                    node.getUserObject().toString().split(" ");

                            if (p[0]
                                    .equals(XMLEncoderUtil.VIEW_POINT_ELEMENTNAME)) {
                                annotationButtons.setSelected(
                                        viewPointRadioButton.getModel(), true);
                            } else if (p[0]
                                    .equals(XMLEncoderUtil.POINT_ATTRNAME)) {
                                double x = Double.parseDouble(p[1]);
                                double y = Double.parseDouble(p[2]);
                                mapRegion.removePoint(x, y);
                            }

                        }
                    }
                } else if (event.equals(AnnotationDataTreePanel.DRIVE2)) {
                    if (!node.isRoot()) {
                        DefaultMutableTreeNode parent =
                                (DefaultMutableTreeNode) node.getParent();

                        if (!parent.isRoot() && !node.isLeaf()) {
                            // region
                            String areaType = node.getParent().toString();
                            String regionLabel =
                                    node.getUserObject().toString();

                            MapArea area = areas.get(areaType);
                            mapRegion = area.getRegionByLabel(regionLabel);
                            
                            try {
                                navActuator.setGoal(mapRegion.getViewPoint());
                            } catch (IOException e) {
                                   LOGGER.warn(e);
                            }

                        } else if (node.isLeaf()) {
                            // point
                            String areaType =
                                    node.getParent().getParent().toString();
                            String regionLabel = node.getParent().toString();

                            MapArea area = areas.get(areaType);
                            mapRegion = area.getRegionByLabel(regionLabel);

                            try {
                                navActuator.setGoal(mapRegion.getViewPoint());
                            } catch (IOException e) {
                                   LOGGER.warn(e);
                            }
                        }
                    }
                }

                updateGuiComponents();
            }
        };
        return listener;
    }

    private static void updateGuiComponents() {
        annotationTreePanel.clearTree();

        for (MapArea entry : areas.values()) {
            annotationTreePanel.addAnnotation(entry);
        }

        mapPanel.updateAnnotations(areas, mapRegion);
    }

    private static WorldClickListener createWorldClickListener(
            final MapAnnotationPanel mapPanel) {
        return new MapAnnotationPanel.WorldClickListener() {

            @Override
            public void worldClicked(Pair<Double> worldCoordinates,
                    MouseEvent mouseEvent) {

                String eventTyp = mouseEvent.paramString().split(",")[0];

                if (eventTyp.equals("MOUSE_CLICKED")) {

                    String annotationCommand =
                            annotationButtons.getSelection().getActionCommand();
                    String label = labelText.getText();
                    double height = Double.parseDouble(heightText.getText());

                    double dx = worldCoordinates.getFirst();
                    double dy = worldCoordinates.getSecond();

                    if (mouseEvent.getButton() == MouseEvent.BUTTON1) {

                        if (mapRegion == null) {
                            mapRegion = new MapRegion(MapRegion.POLYGONE_TYPE);
                        }

                        if (annotationCommand.equals(ANNOTATION)) {

                            mapRegion.addPoint(dx, dy);
                            mapRegion.setLabel(label);
                            mapRegion.setHeight(height);

                        } else {

                            NavigationGoalData viewPoint =
                                    mapRegion.getViewPoint();

                            if (viewPoint == null) {
                                viewPoint =
                                        new NavigationGoalData(
                                                "SlamAnnotationTools", dx, dy,
                                                0, GoalType.GLOBAL);
                                mapRegion.setViewPoint(viewPoint);

                            } else {

                                double offset = 0;

                                double ankat =
                                        Point2D
                                                .distance(mapRegion
                                                        .getViewPoint().getX(),
                                                        mapRegion
                                                                .getViewPoint()
                                                                .getY(), dx,
                                                        mapRegion
                                                                .getViewPoint()
                                                                .getY());
                                double hypo =
                                        Point2D
                                                .distance(mapRegion
                                                        .getViewPoint().getX(),
                                                        mapRegion
                                                                .getViewPoint()
                                                                .getY(), dx, dy);
                                double yaw = Math.acos(ankat / hypo);

                                if (dx > mapRegion.getViewPoint().getX()
                                        && dy > mapRegion.getViewPoint().getY()) {
                                    mapRegion.getViewPoint().setYaw(yaw);
                                } else if (dx < mapRegion.getViewPoint().getX()
                                        && dy > mapRegion.getViewPoint().getY()) {
                                    offset = Math.PI * 0.5;
                                    double dif = offset - yaw;
                                    double yaw2 = dif + offset;
                                    mapRegion.getViewPoint().setYaw(yaw2);
                                } else if (dx < mapRegion.getViewPoint().getX()
                                        && dy < mapRegion.getViewPoint().getY()) {
                                    offset = Math.PI;
                                    double yaw2 = yaw + offset;
                                    mapRegion.getViewPoint().setYaw(yaw2);
                                } else if (dx > mapRegion.getViewPoint().getX()
                                        && dy < mapRegion.getViewPoint().getY()) {
                                    offset = Math.PI + Math.PI * 0.5;
                                    double dif = Math.PI * 0.5 - yaw;
                                    double yaw2 = dif + offset;
                                    mapRegion.getViewPoint().setYaw(yaw2);
                                }

                            }
                        }

                    } else if (mouseEvent.getButton() == MouseEvent.BUTTON2
                            || mouseEvent.getButton() == MouseEvent.BUTTON3) {

                        if (mapRegion == null || mapRegion.getPointCount() < 1) {
                            return;
                        }
                        String areaLabel =
                                areaButtons.getSelection().getActionCommand();

                        String arenaName = arenaText.getText();

                        String regionGroup = groupText.getText();

                        mapRegion.setLabel(label);
                        mapRegion.setGroup(regionGroup);
                        mapRegion.setHeight(height);

                        if (!edit) {
                            if (!areas.containsKey(areaLabel)) {
                                MapArea mapArea =
                                        new MapArea(MapArea.MapAreaType
                                                .valueOf(areaLabel));
                                mapArea.addRegion(mapRegion);
                                mapArea.setArenaName(arenaName);
                                areas.put(areaLabel, mapArea);
                            } else {
                                mapRegion.setLabel(label);
                                mapRegion.setGroup(regionGroup);
                                mapRegion.setHeight(height);
                                areas.get(areaLabel).addRegion(mapRegion);
                            }
                        } else {
                            edit = false;
                        }
                        mapRegion = new MapRegion(MapRegion.POLYGONE_TYPE);

                    }
                    updateGuiComponents();

                }
            }

        };
    }

    private static void saveXML(final String annotationXmlPath) {
        Element root = new Element(ANNOTATION);

        for (MapArea entry : areas.values()) {
            Element area = new Element(AREA);
            XMLEncoderUtil.writeMapArea(area, entry);
            root.appendChild(area);
        }

        try {
            XMLEncoderUtil.xmlWriter(root, annotationXmlPath);
        } catch (IOException ex) {
            LOGGER.warn(ex);
        }
    }

}
