/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.xcf.actuators;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.SyntaxException;
import nu.xom.Document;
import nu.xom.ParsingException;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.CameraActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.SensorListener;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.data.CameraPositionData;
import de.unibi.airobots.btl.data.CameraTargetData;
import de.unibi.airobots.btl.data.PersonData;

/**
 * Actuator which can be used to access the Pan-Tilt-Unit of the Camera using
 * the new java fts camserver.
 * 
 * @author lziegler
 * @author ttoenige
 */
public class XcfCameraActuator implements CameraActuator,
        RemoteServerReconnectable {

    private Logger logger = Logger.getLogger(getClass());

    private String remoteServerName = "";

    private Object remoteServerLock = new Object();
    private RemoteServer remoteServer = null;

    /**
     * Constructor.
     * 
     * @param remoteServerName
     *            name of the remote server used to control the camera
     */
    public XcfCameraActuator(String remoteServerName) {
        this.remoteServerName = remoteServerName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setTarget(CameraTargetData target) throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            Document doc = target.toDocument(CameraTargetData.class);

            synchronized (remoteServerLock) {
                remoteServer.callMethod("setTarget", new XOPData(doc));
            }

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public long getCameraHeight() throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            synchronized (remoteServerLock) {
                XOPData heightResult =
                        remoteServer.callMethod("getCamHeight", XOPData
                                .fromString("<FOO/>"));
                String heightString =
                        heightResult.getDocument().getRootElement()
                                .getAttribute("mm").getValue();
                long height = Long.parseLong(heightString);
                return height;
            }

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public CameraPositionData getPosition() throws IOException {
        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            synchronized (remoteServerLock) {
                XOPData positionResult =
                        remoteServer.callMethod("getPosition", XOPData
                                .fromString("<FOO/>"));

                CameraPositionData pos =
                        CameraPositionData.fromDocument(positionResult
                                .getDocument(), CameraPositionData.class);

                return pos;
            }

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (ParsingException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setPosition(CameraPositionData position) throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            Document doc = position.toDocument(CameraPositionData.class);

            synchronized (remoteServerLock) {
                remoteServer.callMethod("setPosition", new XOPData(doc));
            }

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setSpeed(CameraSpeed speed) throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            String doc = "<CAMERASPEED ROTVEL=\"" + speed.name() + "\"/>";

            synchronized (remoteServerLock) {
                remoteServer.callMethod("setCamSpeed", XOPData.fromString(doc));
            }

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        assert remoteServer.getServerName().equals(remoteServerName);
        synchronized (remoteServerLock) {
            this.remoteServer = remoteServer;
        }
    }

    private int focusPersonId = -1;
    private Sensor<PersonList> personSensor = null;
    private boolean doingCamAction = false;

    private SensorListener<PersonList> personSensorListener =
            new SensorListener<PersonList>() {

                private CameraTargetData camTargetData = new CameraTargetData();

                private PersonData getNearestPerson(PersonList newData) {
                    PersonData nearestPersonData = null;
                    double shortestDistance = Double.MAX_VALUE;

                    for (PersonData p : newData) {
                        if (p.getDistance() < shortestDistance) {
                            nearestPersonData = p;
                            shortestDistance = p.getDistance();
                        }
                    }

                    shortestDistance = Double.MAX_VALUE;
                    return nearestPersonData;
                }

                private PersonData getFocusPersonById(int id, PersonList newData) {
                    PersonData focusPerson = null;
                    for (PersonData p : newData) {
                        if (p.getId() == id) {
                            focusPerson = p;

                        }
                    }
                    return focusPerson;
                }

                private int i = 30;

                @Override
                public void newDataAvailable(PersonList newData) {

                    if (i == 30) {
                        try {
                            if (focusPersonId != -1) {
                                PersonData focusPerson =
                                        getFocusPersonById(focusPersonId,
                                                newData);
                                if (focusPerson != null) {
                                    camTargetData.setDistance(focusPerson
                                            .getDistance());
                                    camTargetData.setHeight(1700);
                                    camTargetData.setAzimuth(focusPerson
                                            .getAngle());
                                    setTarget(camTargetData);
                                    logger.debug(camTargetData);
                                }
                            } else {
                                PersonData nearestPersonData =
                                        getNearestPerson(newData);
                                if (nearestPersonData != null) {
                                    camTargetData.setDistance(nearestPersonData
                                            .getDistance());
                                    camTargetData.setHeight(1700);
                                    camTargetData.setAzimuth(nearestPersonData
                                            .getAngle());

                                    setTarget(camTargetData);

                                    logger.debug(camTargetData);
                                }
                            }

                        } catch (IOException ex) {
                            logger.error(ex);
                        }
                        i = 0;
                    }
                    i++;

                }

            };

    /**
     * {@inheritDoc}
     */
    @Override
    public void startCamAction(Sensor<PersonList> personSensor) throws IOException {
        // TODO what happened if person go out of the laser?!
        doingCamAction = true;
        this.personSensor = personSensor;
        this.personSensor.addSensorListener(personSensorListener);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void stopCamAction() throws IOException {
        doingCamAction = false;
        if (personSensor != null && personSensorListener != null) {
            personSensor.removeSensorListener(personSensorListener);
            personSensor = null;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void focus(Sensor<PersonList> personSensor, int personId)
            throws IOException {
        // TODO NullPointer abfangen?! Was ist wenn die Person nicht da ist
        this.personSensor = personSensor;
        this.personSensor.removeSensorListener(personSensorListener);
        this.personSensor.addSensorListener(personSensorListener);
        focusPersonId = personId;

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void unfocus() throws IOException {
        if (!doingCamAction && personSensor != null
                && personSensorListener != null) {
            this.personSensor.removeSensorListener(personSensorListener);
        }
        focusPersonId = -1;

    }
}
