/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.xcf.actuators;

import java.io.IOException;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.memory.DocumentNotFoundException;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.XPath;
import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.RobotStateActuator;
import de.unibi.airobots.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.btl.data.RobotState;
import de.unibi.airobots.btl.data.Type.NoSourceDocumentException;

/**
 * @author ltwardon
 */
public class XcfRobotStateActuator implements RobotStateActuator,
        ActiveMemoryReconnectable {

    private Logger logger = Logger.getLogger(getClass());

    private String memoryName = "";
    private Object memoryLock = new Object();
    private ActiveMemory memory = null;
    private boolean isInserted = false;
    private XOPData original;
    
    /**
     * Xml namespace for active memory documents.
     */
    public static final String NAMESPACE_MEMORY =
            "http://purl.org/ActiveMemory";

    /**
     * Xml namespace for dbxml.
     */
    public static final String NAMESPACE_DBXML =
            "http://www.sleepycat.com/2002/dbxml";

    /**
     * Creates a new instance of {@link XcfRobotStateActuator}.
     * 
     * @param memoryName
     *            Name of the memory to connect to.
     */
    public XcfRobotStateActuator(String memoryName) {
        this.memoryName = memoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getActiveMemoryName() {
        return memoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {
        synchronized (memoryLock) {
            this.memory = activeMemory;
        }
    }

    @Override
    public void sendState(RobotState state) throws IOException {

        if (memory == null) {
            logger.error("Memory is not reachable");
        }
        try {
            if(isInserted) {
                Element source = original.getDocument().getRootElement();
                state.fillInto(source);
                logger.info("Replace: " + source.getDocument().toXML());
                memory.replace(new XOPData(source.getDocument()));
            }
            else {
                original = memory.insert(new XOPData(state.toDocument(RobotState.class)));
                logger.info("Insert: " + original.getDocument().toXML());
                isInserted = true;
            }
        } catch (DocumentNotFoundException e) {
            logger.error("Document not found in memory");
            logger.debug(e);
            throw new IOException(e);
        } catch (MemoryException e) {
            logger.error("Memory is not reachable");
            logger.debug(e);
            throw new IOException(e);
        }
    }

}
