package de.unibi.airobots.bonsai.xcf.configuration;

import java.util.Map;

import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.CoreObjectFactory.SensorToConfigure;

/**
 * Base class for XCF-based sensor configuration objects that use a memory 
 * instead of a subscriber to a publisher. The default
 * implementation of {@link #parseOptions(Map)} requires a memory name for
 * convenience reasons.
 * 
 * @author lziegler
 * @param <T>
 *            data type returned by the sensor
 */
public abstract class ConfiguredXcfMemorySensor<T> extends ConfiguredXcfSensor<T> {

    /**
     * Option for publisher based sensors defining the memory name.
     */
    public static final String OPTION_MEMORY_NAME = "memoryName";

    /**
     * Data type to be returned by the sensor.
     */
    private Class<T> dataType = null;

    /**
     * Name of the configured memory if {@link #parseOptions(Map)} was used.
     */
    private String memoryName = "";

    /**
     * Create a new specific sensor instance.
     * 
     * @return new instance
     */
    public abstract Sensor<T> createInstance();

    /**
     * {@inheritDoc}
     * 
     * Default implementation parsing the publisher name.
     */
    @Override
    public void parseOptions(Map<String, String> options)
            throws IllegalArgumentException {

        if (!options.containsKey(OPTION_MEMORY_NAME)) {
            throw new IllegalArgumentException(
                    "Memory name missing in arguments.");
        }

        memoryName = options.get(OPTION_MEMORY_NAME);

    }

    /**
     * Override this and check for suitable data type!
     * 
     * @param sensor
     *            requested sensor
     * @return <code>true</code> if configuration can create such a requested
     *         sensor
     */
    public static boolean isSuitableFor(SensorToConfigure sensor) {
        return false;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String[] getActiveMemoriesToInitialize() {
        return new String[] { memoryName };
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public String[] getSubscribersToInitialize() {
        return new String[] { };
    }

    /**
     * Sets the data type the configured sensor is intended to return.
     * 
     * @param dataType
     *            new data type
     */
    @Override
    public void setDataType(Class<T> dataType) {
        this.dataType = dataType;
    }

    /**
     * Returns the data type the configured sensor will return.
     * 
     * @return data type class
     */
    @Override
    public Class<T> getDataType() {
        return dataType;
    }

    /**
     * Returns the configured memory name for the sensor if
     * {@link #parseOptions(Map)} was used.
     * 
     * @return publisher name
     */
    public String getMemoryName() {
        return memoryName;
    }

}
