/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.xcf.sensors;

import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.event.MemoryEvent;
import net.sf.xcf.event.MemoryEventAdapter;
import net.sf.xcf.event.MemoryEventListener;
import net.sf.xcf.memory.MemoryAction;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.xml.XPath;
import nu.xom.ParsingException;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.core.SensorListener;
import de.unibi.airobots.bonsai.sensors.PersonSensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.btl.data.PersonData;

/**
 * The person sensor receives the person list provided by the PTA (MMAnchoring).
 * The sensor provides the list with all persons currently in the robots memory.
 * 
 * TODO:
 * <ul>
 * <li>Maybe notifying listeners needs to be filtered so that listeners are only
 * notified if there is <b>a</b> change in the data.</li>
 * </ul>
 * 
 * @author lschilli
 * @author jwienke
 */
public class XcfActiveMemoryPersonSensor implements ActiveMemoryReconnectable,
        PersonSensor {

    private Logger logger = Logger.getLogger(getClass());

    private Object sharedDataLock = new Object();

    /**
     * Map of {@link PersonData} indexed by their id. Always acquire
     * {@link #sharedDataLock} before using this map.
     */
    private Hashtable<Integer, PersonData> personmap =
            new Hashtable<Integer, PersonData>();

    /**
     * TODO what exactly is this? I'm sure this also needs locking via
     * {@link #sharedDataLock}.
     */
    private LinkedList<PersonData> currentlySeenList =
            new LinkedList<PersonData>();

    // private PositionSensor positionSensor;

    /**
     * Own listener used to intercept the memory for person changes.
     */
    private MemoryEventListener selfListener;

    /**
     * Set of sensor listeners. Only access this set after acquiring
     * {@link #sharedDataLock}.
     */
    private Set<SensorListener<PersonList>> listeners =
            new HashSet<SensorListener<PersonList>>();

    private String memoryName = "";
    private Object memoryLock = new Object();
    private ActiveMemory memory = null;

    /**
     * Constructor.
     * 
     * @param memoryName
     *            name of the active memory to get the persons from
     */
    public XcfActiveMemoryPersonSensor(String memoryName) {

        this.memoryName = memoryName;

        logger.debug("Create memory person Sensor for memoryName " + memoryName);
        
        // TODO why was this needed? How to manage this case?
        // positionSensor =
        // BonsaiFactory.getPositionSensorManager().getSensor();

        selfListener =
                new MemoryEventAdapter(MemoryAction.ALL, new XPath("/PERSON")) {

                    @Override
                    public void handleEvent(MemoryEvent e) {

                        
                        //logger.debug("new Person Event  " + e.getType());
                        try {

                            // TODO what about REPLACE here?

                            if (e.getType() == MemoryAction.INSERT
                                    || e.getType() == MemoryAction.REPLACE) {

                                PersonData p =
                                        PersonData.fromDocument(e.getData()
                                                .getDocument(),
                                                PersonData.class);

                                synchronized (sharedDataLock) {

                                   // logger.debug("person ID " + p.getId());
                                    
                                    personmap.put(p.getId(), p);

                                    try {
                                        PersonList list =
                                                getPersonListFromMemory();
                                        for (SensorListener<PersonList> l : listeners) {
                                            l.newDataAvailable(list);
                                        }
                                    } catch (IOException ex) {
                                        logger
                                                .error(
                                                        "Error notifying observers",
                                                        ex);
                                    }

                                }

                            } else if (e.getType() == MemoryAction.REMOVE) {

                                PersonData p =
                                        PersonData.fromDocument(e.getData()
                                                .getDocument(),
                                                PersonData.class);

                                synchronized (sharedDataLock) {

                                    personmap.remove(p.getId());

                                    try {
                                        PersonList list =
                                                getPersonListFromMemory();
                                        for (SensorListener<PersonList> l : listeners) {
                                            l.newDataAvailable(list);
                                        }
                                    } catch (IOException ex) {
                                        logger
                                                .error(
                                                        "Error notifying observers",
                                                        ex);
                                    }

                                    Iterator<PersonData> it =
                                            currentlySeenList.iterator();
                                    boolean seen = false;
                                    while (it.hasNext()) {

                                        if (p.getId() == it.next().getId()) {
                                            seen = true;
                                        }

                                    }

                                    if (currentlySeenList.size() <= PERSON_REMEMBER_LIMIT
                                            && !(seen)) {
                                        logger.debug("Pushing person number: "
                                                        + p.getId());
                                        currentlySeenList
                                                .push(deriveGlobalPosition(p));
                                    } else if (currentlySeenList.size() > PERSON_REMEMBER_LIMIT
                                            && !(seen)) {
                                        currentlySeenList.pop();
                                        logger.debug("First poping now pushing person number: "
                                                        + p.getId());
                                        currentlySeenList.push(p);
                                    }

                                }

                            }

                        } catch (ParsingException ex) {
                            logger.error(ex);
                        }

                    }

                    /**
                     * This function calculates the current global position of a
                     * person according to the observed odometry data from the
                     * robot.
                     * 
                     * @param p
                     *            Person
                     * @return the inserted person with a global position.
                     */
                    private PersonData deriveGlobalPosition(PersonData p) {

                        // TODO why was this doing nothing?

                        // try {
                        //
                        //
                        // PositionData posData = positionSensor.readLast();
                        // NavigationGoalData navGoal =
                        // CoordinateSystemConverter.polar2NavigationGoalData(posData,
                        // p.getAngle(), p.getDistance());
                        // posData.setX(navGoal.getX());
                        // posData.setY(navGoal.getY());
                        // posData.setY(navGoal.getYaw());
                        //
                        // //p.setPositionData(posData);
                        //
                        //
                        // } catch (IOException e) {
                        // e.printStackTrace();
                        // }

                        return p;
                    }
                };
                
                logger.debug("End memory person Sensor for memoryName ");

    }

    private PersonList getPersonListFromMemory() throws IOException {
        PersonList pl = new PersonList();
       // logger.debug("hasNext " + hasNext());
        synchronized (sharedDataLock) {
             Collection<PersonData> it = personmap.values();
             for (PersonData p : it) {
         //        logger.debug("person in list " + p.getId());
                 pl.addPerson(p);
             }
        }
        return pl;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public PersonList readLastSeenPersons() {
        PersonList pl = new PersonList();
        synchronized (sharedDataLock) {
            Iterator<PersonData> it = currentlySeenList.iterator();
            while (it.hasNext()) {
                pl.addPerson(it.next());
            }
        }
        return pl;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void addSensorListener(SensorListener<PersonList> listener) {
        synchronized (sharedDataLock) {
            listeners.add(listener);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Class<PersonList> getDataType() {
        return PersonList.class;
    }

    /**
     * Delivers a list of the last persons tracked by the sensor.
     * 
     * {@inheritDoc}
     */
    @Override
    public PersonList readLast() throws IOException {
        logger.debug("readlast");
        return getPersonListFromMemory();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public PersonList readLast(long timeout) throws IOException,
            InterruptedException {
        return readLast();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void removeSensorListener(SensorListener<PersonList> listener) {
        synchronized (sharedDataLock) {
            listeners.remove(listener);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getActiveMemoryName() {
        return memoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {

        synchronized (memoryLock) {
            this.memory = activeMemory;
        }

        try {
            memory.addListener(selfListener);
        } catch (MemoryException e) {
            logger.error("Error adding listeners to new memory instance", e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean hasNext() {
        return personmap.size() > 0;
    }

}
